--  BYTE-compare of two or more DIRs.    Ver. 1.2     Rolf Schrder 1999-10-20
--                   Mrkenweg 37
--                                                    21029 Hamburg , Germany
--  Language: EUPHORIA ver. 2.1                       r.schroeder@bergedorf.de
--
--  To account for long filenames: copy or rename this file with the ending
--                 
--  *.EXW and run it with EXW or bind it with BINDW to run it directly.
--
------------------------------------------------------------------------------
--
--  When burning CD-ROMs under WIN 3.11 I had the possibility to verify the
--  copy with the orginal CD or HD-dir.  Sometimes I encountered byte errors.
--  Unfortunately, a new software I'm using now under Win NT has no more the
--  verify option.
--
--  DIRCMP was written to compare the burned CD-ROM with its original data
--  source which might be an CD-ROM or a hard disks dir. It runs with EX under
--  DOS or with EXW under Win NT to account for long filenames.
--
--  No FileName-, FileLength-, DirContent-, or FileContent- mismatches are
--  accepted. Only 10 Byte- mismatches are allowed before the program aborts.
--
--  Please report any problem or error and I will try to help you.
--
------------------------------------------------------------------------------
without type_check

include file.e              -- open(), close(), dir(), seek()
include sort.e              -- sort()
include get.e               -- wait_key()
include graphics.e          -- get_position()
include misc.e              -- platform()
------------------------------------------------------------------------------
-- return strings for function dir_search():

constant ret1 = "Less then 2 paths in dir_search! Abort.\n"
constant ret2 = "%s does not exist (or is an empty ROOT)! Abort.\n"
constant ret3 = "%s does not exist! Abort.\n"
constant ret4 = "File content of DIRs %s and %s are different! Abort.\n"
constant ret5 = "Different filenames in %s and %s! Abort.\n"
constant ret6 = "Lengths of files %s and %s are different! Abort.\n"
constant ret7 = "Unable to open %s! Abort.\n"
constant ret8 = "More then 10 BYTE-mismatches encountered! Abort.\n"
constant ret9 = "%s and %s are different at BYTE %d!\n"
constant retA = "Aborted by user.\n"
------------------------------------------------------------------------------

global integer  errors      -- Error count
global integer  files       -- File count
global sequence init_cuspos -- Initial cursor position for activity_mark()

------------------------------------------------------------------------------

procedure activity_mark( )  -- needs global sequence init_cuspos
--        
    sequence keep_cuspos

    keep_cuspos = get_position()            -- position of 'normal' output
    if init_cuspos[2] <= 80 then            -- add one mark to line
      position ( init_cuspos[1], init_cuspos[2])  -- pos. for next mark
    else                                    -- restart line and set one mark
      init_cuspos[2] = 1
      position ( init_cuspos[1], 1)
      puts( 1, repeat( ' ', 80) )           -- clear line
      position ( init_cuspos[1], 1)
    end if
    puts( 1,"*")                            -- set mark
    init_cuspos[2] += 1                     -- set next cuspos
    position( keep_cuspos[1], keep_cuspos[2])  -- return to initial cuspos

end procedure

------------------------------------------------------------------------------

function dir_search( sequence path )        -- Recursive procedure
--       
    object   x                              -- return from dir_search()
    sequence dir_c, sjunk                   -- "content" of DIRs, junk
    integer  fn1, fn2, ijunk                -- FILE-IDs, junk

    if length(path) < 2 then                -- Test for number of paths
      return( ret1 )
    end if
    for i = 1 to length(path) do            -- Check for '\\' at end of paths
      if path[i][length(path[i])] != '\\' then -- if end of path not '\\' then
	path[i] = path[i] & '\\'            --  concatenate '\\'.
      end if
    end for
    dir_c = repeat({},length(path))
    for i = 1 to length(path) do        -- Go through paths
      dir_c[i] = dir(path[i])
      if atom(dir_c[i]) then            -- If one DIR does't exist...
	if length(path[i]) < 4 then
	  return( sprintf( ret2, {path[i]} ))
	else
	  return( sprintf( ret3, {path[i]} ))
	end if
      end if
      ijunk = 0
      for j = 1 to length(dir_c[i]) do  -- Are there "."- or ".."-DIRs?
	if dir_c[i][j][1][1] = '.' then -- It's not a ROOT-DIR
	  ijunk = 1
	  exit
	end if
      end for
      if ijunk = 0 then                 -- Add "." and ".." -DIRs.
	dir_c[i] = prepend( dir_c[i], {"..","d",0,0,0,0,0,0,0}) -- Dummy-DIRs
	dir_c[i] = prepend( dir_c[i], {"." ,"d",0,0,0,0,0,0,0}) -- (for ROOTs)
      end if
    end for
    for i = 2 to length(path) do        -- Check for equal contents of DIRs
      if length(dir_c[1]) != length(dir_c[i]) then -- DIR-contents different!
	return( sprintf( ret4, {path[1],path[i]} ) )
      end if
    end for
    for i = 1 to length(path) do        -- Sort for compare.
      dir_c[i] = sort( dir_c[i] )
    end for
    for i1 = 1 to length(dir_c[1]) do   -- Compare File- Names and -Length.
      for i2 = 2 to length(path) do
	if compare(dir_c[1][i1][1],dir_c[i2][i1][1]) then -- FileNames equal?
	  return( sprintf( ret5, {path[1], path[i2]}) )
	end if
	if dir_c[1][i1][3] != dir_c[i2][i1][3]      -- Test FileLength, but
	 and not find( 'd', dir_c[1][i1][2]) then   -- not for DIR-Names!
	  return(sprintf( ret6, {path[ 1] & dir_c[ 1][i1][1],
				 path[i2] & dir_c[i2][i1][1]}) )
	end if
      end for
    end for
    for i1 = 1 to length(dir_c[1]) do     -- Loop over DIR-content!
      if find( 'd', dir_c[1][i1][2]) then -- SUB-DIR found.
	if dir_c[1][i1][1][1] != '.' then -- It's not an OWN- or UP-DIR, go
	  sjunk = path                    --   into next SUB-DIR!
	  for j = 1 to length(path) do    -- Add path's to SUB-DIRs.
	    sjunk[j] = path[j] & dir_c[1][i1][1] & '\\'
	  end for
	  x = dir_search( sjunk )         -- Recursiv call of dir_search().
	  if sequence(x) then             -- In case of error return with
	    return(x)                     -- sequence else with atom.
	  end if
	end if
      else                                -- handle FILE
	if dir_c[1][i1][3] > 0 then       -- Compare only if FileLenght > 0.
	  fn1 = open( path[1] & dir_c[1][i1][1], "rb" )
	  if fn1 = -1 then
	    return( sprintf( ret7, {path[1] & dir_c[1][i1][1]}) )
	  end if
	  for i3 = 2 to length(path) do   -- LOOP over FILES to compare.
	    files += 1
	    activity_mark()
	    ijunk = seek(fn1,0)           -- Startposition of origin-FILE.
	    fn2 = open( path[i3] & dir_c[1][i1][1], "rb" )
	    if fn2 = -1 then
	      close( fn1 )
	      return( sprintf( ret7, {path[i3] & dir_c[1][i1][1]}) )
	    end if
	    for loop = 1 to dir_c[1][i1][3] do  -- LOOP over BYTES to compare.
	      if getc(fn1) != getc(fn2) then
		errors += 1
		if errors > 10 then
		  close( fn1 )
		  close( fn2 )
		  return( ret8 )
		end if
		printf( 1, ret9, {path[ 1] & dir_c[1][i1][1],
				  path[i3] & dir_c[1][i1][1], loop})
	      end if
	    end for
	    close( fn2 )
	    if get_key() = 27 then  -- abort if ESC was pressed...
	      close( fn1 )
	      return( retA )
	    end if
	  end for
	  close( fn1 )
	end if
      end if
    end for
    return 0
end function

------------------------------------------------------------------------------

procedure main( sequence cmd )
--        
    object x                        -- return from dir_search(), wait_key()
    sequence dt1, path              -- DATE, Paths
    sequence cuspos

    if length(cmd) < 4 then         -- Show how to use it.
      wrap(0)                       -- wrap off
      puts( 1, "Ŀ\n")
      puts( 1, "                                                                              \n")
      puts( 1, " Compare all FILES within a DIR-structure (origin) with all FILES of other    \n")
      puts( 1, " DIR-structures (replica).                                                    \n")
      puts( 1, "                                                                              \n")
      puts( 1, " No FileName-, FileLength-, DirContent-, or FileContent- mismatches are       \n")
      puts( 1, " accepted. However, 10 Byte- mismatches will be monitored before the          \n")
      puts( 1, " program aborts.                                                              \n")
      puts( 1, "                                                                              \n")
      puts( 1, " Example:                                                                     \n")
      puts( 1, "                                                                              \n")
      if compare(cmd[1],cmd[2]) then
	puts( 1, "   " & cmd[1] & " " & cmd[2] & " C: K:\\C_COPY\\ D:\\TST")
      else
	puts( 1, "   " & cmd[2] & " C: K:\\C_COPY\\ D:\\TST")
      end if
	cuspos = get_position()
	position( cuspos[1], 80) puts( 1, "\n")
      puts( 1, "                                                                              \n")
      puts( 1, "   All files in drive C:\\ will be compared byte-wise with all corresponding   \n")
      puts( 1, "   files in K:\\C_COPY\\ and D:\\TST\\                                            \n")
      puts( 1, "                                                                              \n")
      puts( 1, " Press ESC to abort program.                                                  \n")
      puts( 1, "                                                                              \n")
      puts( 1, "\n")
      wrap(1)                       -- wrap on
      if platform() = WIN32 then
	puts( 1, "\n  exit with any key...")
	x = wait_key()
      end if
      abort(0)
    end if
    clear_screen()
    puts( 1, cmd[2] & " compares the following DIRs:\n")
    path = {}
    for i = 3 to length(cmd) do
      path = append( path, cmd[i] )     -- Origin- und Replica-StartDirs
      printf( 1, "DIR %d: %s\n", { i-2, path[i-2]})
    end for
    errors = 0      -- Error count in dir_search()
    files  = 0      -- FILE  count in dir_search()
    dt1 = date()  dt1[1] += 1900
    printf( 1, "Start: %04d-%02d-%02d %02d:%02d:%02d (Press ESC to abort program)\n", dt1)
    init_cuspos = get_position()        -- keep cuspos for activity_mark()
    puts( 1, "-\n")                     -- initialize actiyity output
    x = dir_search( path )
    if sequence(x) then
      puts( 1, x)
    elsif errors > 0 then
      printf( 1, "BYTE-errors: %d (max. 10)\n", errors)
      printf( 1, "%d FILES (replica) compared.\n", files)
    else
      printf( 1, "%d FILES (replica) compared, no error encountered.\n", files)
    end if
    dt1 = date()  dt1[1] += 1900
    printf( 1, "Stop : %04d-%02d-%02d %02d:%02d:%02d\n", dt1)
    if platform() = WIN32 then
      puts( 1, "\n  exit with any key...")
      x = wait_key()
    end if
end procedure

------------------------------------------------------------------------------
--
--  Start (and Test) of main.
--
    constant TEST = 0                   -- Set to 1 for testing under ED.EX .
    constant dir1 = "c:\\acrobat3"      -- DIR1 to compare under ED.EX with
    constant dir2 = "k:\\acrobat3"      -- DIR2 in case TEST != 0.

    sequence cmd
    cmd = command_line()

    if TEST and compare(cmd[1],cmd[2]) then
      cmd = append( cmd, dir1 )
      cmd = append( cmd, dir2 )
    end if

    main(cmd)
    abort(0)
--
------------------------------------------------------------------------------
